from dataclasses import dataclass, field
from datetime import datetime
from typing import Any
from uuid import uuid4

from thescientist.dataset import Dataset


# dependent variable
@dataclass(frozen=True)
class Metric:
    name: str
    dtype: type
    not_null: bool = field(default=True)

    def as_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "dtype": str(self.dtype.__name__),
            "not_null": self.not_null
        }


class MetricCollector:
    def __init__(self, name, metrics: list[Metric]) -> None:
        self.name = name
        self.original_metrics = metrics
        self.metrics = {
            metric.name: {
                "schema": metric,
                "dataset": Dataset.empty(schema={"id": str, metric.name: metric.dtype, "collected_at": str})
            }
            for metric in metrics
        }
    
    def __repr__(self) -> str:
        return f"Collector(metrics={self.metrics})::data =\n{self.datasets}"

    def collect(self, metrics: dict[str, Any]) -> None:
        collected_id = str(uuid4())
        collected_at = datetime.now().isoformat()  # datetime.now(timezone.utc)
        
        for metric_name, value in metrics.items():
            metric = self.metrics.get(metric_name)

            if not metric:
                print(f"[WARNING] metric '{metric_name}' not defined.")
                continue

            data = {"id": collected_id, metric["schema"].name: value, "collected_at": collected_at}

            if (value is None) and metric["schema"].not_null:
                print(f"[WARNING] metric '{metric_name}' can not be null.")
                continue

            metric["dataset"].append(data)



import unittest

# from unittest import TestCase

from science.dataset import Dataset
from science.metric import Metric, MetricCollector


class MetricCollectorTest(unittest.TestCase):

    def test_collect(self):
        
        collector = MetricCollector(name="network-metrics",
                                    metrics_schema=[Metric("waiting_time", float),
                                                    Metric("total_packages", int),
                                                    Metric("package_type", str)])
        
        collector.collect({"waiting_time": 1.1,
                           "total_packages": 2})
        
        collector.collect({"waiting_time": 1.2,
                           "package_type": "UDP"})
        
        print(f"dset = \n{collector.metrics}")

        result = collector.metrics.data
        expected = Dataset.from_dict({"waiting_time": [1.1, 1.2], "total_packages": [2, None], "package_type": [None, "UDP"]}).data

        self.assertListEqual(result["waiting_time"].to_list(), expected["waiting_time"].to_list())
        # self.assertListEqual(result["total_packages"].to_list(), expected["total_packages"].to_list())
        self.assertListEqual(result["package_type"].to_list(), expected["package_type"].to_list())        


if __name__ == "__main__":
    unittest.main(MetricCollectorTest(), verbosity=3)




from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
import json
import os
from typing import Any, Callable, Optional
import uuid
# from thescientist.analysis import AnalyticalData
from thescientist.dataset import Dataset
from thescientist.metric import Metric, MetricCollector


class ExperimentalGroup:
    
    def __init__(self, name: str, algorithm: Optional[Callable[(...), Any]] = None, metadata: Optional[dict[str, Any]] = None) -> 'ExperimentalGroup':
        if not metadata:
            metadata = dict()
        
        self.name = name
        self.algorithm = algorithm
        self.metadata = {"name": name, **metadata}

    def __repr__(self) -> str:
        return f"ExperimentalGroup(name={self.name}, algorithm={self.algorithm}, metadata={self.metadata})"

    @classmethod
    def load_from(cls, file_path: str, algorithm: Optional[Callable[(...), Any]] = None) -> 'ExperimentalGroup':
        with open(file_path, mode="rb") as fp:
            metadata = json.load(fp)
            return cls(name=metadata.get("name", str(uuid.uuid4())),
                       algorithm=algorithm,
                       metadata=metadata)

    def save(self, path_to_directory: str) -> None:
        
        metadata = self.metadata
        metadata["saved_at"] = datetime.now()

        directory = path_to_directory[:-1] if path_to_directory[:-1] == "/" else path_to_directory
        file_path = f"{directory}/{self.name}.group.sci"

        with open(file_path, mode="w") as fp:
            json.dump(metadata, fp)
    
    def run_with(self, *args, **kwargs) -> Any:
        return self.algorithm(*args, **kwargs)


@dataclass
class ExperimentalResult:
    group: ExperimentalGroup
    metric: Metric
    dataset: Dataset
    
    # def to_analytical_data(self) -> AnalyticalData:
    #     return AnalyticalData(metric_name=self.metric.name,
    #                           group_name=self.group.name,
    #                           data=self.dataset.col(self.metric.name))


class Experiment:

    def __init__(self, 
                 name: str, 
                 groups: list[ExperimentalGroup], 
                 metrics: list[Metric], 
                 experimental_function: Callable[[ExperimentalGroup, MetricCollector], None],
                 n_executions: int = 31, 
                 metadata: Optional[dict[str, Any]] = None) -> None:
        
        if not metadata:
            metadata = dict()
        
        self.name = name
        self.groups = groups
        self.metric_collector = MetricCollector(name=name, metrics=metrics)
        self.experimental_function = experimental_function
        self.n_executions = n_executions
        self.metadata = {"name": name, 
                         "groups": [group.name for group in groups], 
                         "metrics": [metric.as_dict() for metric in self.metric_collector.original_metrics],
                         # "experimental_function": experimental_function, 
                         "n_executions": n_executions,
                         **metadata}

        self.results: list[ExperimentalResult] = []
    
    def run_with(self, independent_variables: Any, *args, **kwargs) -> None:
        group_collector = [(group, deepcopy(self.metric_collector)) for group in self.groups]
        for epoch in range(self.n_executions):
            for group, collector in group_collector:
                
                self.experimental_function(group=group, collector=collector, independent_variables=independent_variables, *args, **kwargs)
        
        for group, collector in group_collector:
            for _, metric in collector.metrics.items():
                self.results.append(ExperimentalResult(group=group, metric=metric["schema"], dataset=metric["dataset"]))
    
    def save(self, path_to_directory: str) -> None:
        dt = str(datetime.now()).replace(" ", "_").replace(":", "-").replace(".", "-").replace("+", "-")

        directory = path_to_directory[:-1] if path_to_directory[:-1] == "/" else path_to_directory
        path = f"{directory}/{self.name}/{dt}"

        os.makedirs(path, exist_ok=True)
        
        for result in self.results:
            result.dataset.to_csv(f"{path}/{result.group.name}_{result.metric.name}.data.csv")

        with open(f"{path}/.metadata.experiment.sci", mode="w") as fp:
            json.dump(self.metadata, fp)


if __name__ == "__main__":

    def experiment_function(group: ExperimentalGroup, collector: MetricCollector, independent_variables: list[Any]) -> None:
        result = group.run_with(independent_variables)
        collector.collect({"abs_value": result})
        collector.collect({"rel_value": result})

    def group1(x: int): return 2*x + 1
    def group2(x: int): return x**2 + x + 1
    def group3(x: int): return 3*x**2 - 5*x
    
    experiment = Experiment(name="Absolute-Value", 
                            groups=[ExperimentalGroup(name="decision-tree", algorithm=group1),
                                    ExperimentalGroup(name="svm", algorithm=group2),
                                    ExperimentalGroup(name="mlp", algorithm=group3)],
                            metrics=[Metric("abs_value", int),
                                     Metric("rel_value", int)],
                            experimental_function=experiment_function,
                            n_executions=3)

    experiment.run_with(independent_variables=2)
    experiment.save("./results")




from dataclasses import dataclass, field
from typing import Any, Union
import pandas as pd


Array = pd.Series
DataType = pd.DataFrame


@dataclass(frozen=True)
class Schema:
    name: str
    dtype: type
    not_null: bool = field(default=True)


class Dataset:
    def __init__(self, data: DataType, schema: dict[str, type] = None) -> None:
        self.data = data
        self.schema = schema if schema else dict() 
    
    def __repr__(self):
        return f"{self.data}"
    
    @classmethod
    def from_csv(cls, path: str) -> 'Dataset':
        dataset: DataType = pd.read_csv(path)
        return cls(dataset)
    
    @classmethod
    def from_dict(cls, data: dict[str, list]) -> 'Dataset':
        dataset: DataType = pd.DataFrame(data)
        return cls(dataset)

    @classmethod
    def empty(cls, schema: dict[str, type]) -> 'Dataset':
        # dataset: DataType = pd.DataFrame(columns=list(schema.keys()))
        dataset: DataType = pd.DataFrame({name: pd.Series(dtype=dtype) for name, dtype in schema.items()})
        return cls(dataset, schema)
    
    def to_csv(self, path: str) -> None:
        self.data.to_csv(path, header=True, index=False)
    
    def to_parquet(self, path: str) -> None:
        self.data.to_parquet(path, compression="brotli", header=True, index=False)
    
    def col(self, name: str) -> Array:
        return self.data[name]

    def append(self, data: dict[str, Any]) -> None:
        new_data = pd.DataFrame({name: pd.Series(data=data[name], dtype=dtype) for name, dtype in self.schema.items()})
        if len(self.data) > 0:
            self.data = pd.concat([self.data, new_data], ignore_index=True)
        else:
            self.data = new_data

    def normalize(self, cols: list[str]) -> 'Dataset':
        new_data = self.data.copy()

        for col in cols:
            x = new_data[col]
            x_min = x.min()
            x_max = x.max()
            
            new_data[col] = (x - x_min)/(x_max - x_min)

        return Dataset.from_dict(data=new_data)

    def standardize(self, cols: list[str]) -> 'Dataset':
        new_data = self.data.copy()

        for col in cols:
            x = new_data[col]
            u = x.mean()
            s = x.std()
            
            new_data[col] = (x - u)/s

        return Dataset.from_dict(data=new_data)

